/****************************************************************************
* wanpipe_bsd_iface.c	WANPIPE(tm) OpenSource WANPIPE Driver.
*
* Author:		Alex Feldman	<al.feldman@sangoma.com>
*		
*
* Copyright:		(c) 2004 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 28, 2004	Alex Feldman	Initial version.
*****************************************************************************/

#if defined(__NetBSD__) || defined (__FreeBSD__) || defined(__OpenBSD__)

#include <net/wanpipe_version.h>
#include <net/wanpipe_includes.h>
#include <net/sdlapci.h>
#include <net/wanpipe.h>	/* WANPIPE common user API definitions */
#include <net/if_wanpipe_common.h>
#include <net/wanpipe_iface.h>

#else

# error "Ivalid Operating System!"

#endif

/****** Defines & Macros ****************************************************/

#ifdef	_DEBUG_
#define	STATIC
#else
#define	STATIC		static
#endif

/****** Function Prototypes *************************************************/

static netdevice_t*	wan_iface_alloc (int);
static void		wan_iface_free (netdevice_t*);
static int		wan_iface_attach(netdevice_t*, char*, int);
static void		wan_iface_detach(netdevice_t*, int);
static int		wan_iface_input(netdevice_t*, netskb_t*);
static int 		wan_iface_set_proto(netdevice_t*, struct ifreq*);

#if defined(__FreeBSD__)
static void		wan_iface_init(void*);
#endif
static int		wan_iface_ioctl(netdevice_t*, u_long, caddr_t);
static void		wan_iface_watchdog(netdevice_t*);
static int		wan_iface_output(netdevice_t*, netskb_t*, struct sockaddr*, struct rtentry *);
#if 0
static void		wan_iface_start(netdevice_t*);
#endif

/****** Global Data ********************************************************/
wan_iface_t wan_iface = 
{
	wan_iface_alloc,	/* alloc */
	wan_iface_free,		/* free */
	wan_iface_attach,	/* attach */	
	wan_iface_detach,	/* detach */	
	wan_iface_input,	/* input */	
	wan_iface_set_proto	/* set_proto */	
};

/******* WAN Device Driver Entry Points *************************************/

static netdevice_t* wan_iface_alloc (int is_netdev)
{
	void	*dev;

	if (!is_netdev && !WAN_SPPP_ENABLED){
		DEBUG_EVENT("Please verify that SPPP is enabled in your kernel!\n");
		return NULL;
	}

	/* Register in HDLC device */
	dev = wan_malloc((is_netdev) ? sizeof(struct ifnet) : sizeof(struct sppp));
	if (dev == NULL){
		DEBUG_EVENT("Failed allocate interface structure !\n");
		return NULL;
	}

	bzero(dev, (is_netdev) ? sizeof(struct ifnet) : sizeof(struct sppp));
	return (netdevice_t*)dev;
}

static void wan_iface_free (netdevice_t* dev)
{
	wan_free(dev);
}

static int wan_iface_attach (netdevice_t* dev, char *ifname, int is_netdev)
{
	if (ifname){
		wan_netif_init(dev, ifname);
	}
	dev->if_mtu		= PP_MTU;
	dev->if_flags		= IFF_POINTOPOINT;
#if defined(__FreeBSD__)
	dev->if_init		= wan_iface_init;/* Will set from new_if() */
#endif
	dev->if_ioctl		= wan_iface_ioctl;/* Will set from new_if() */
	dev->if_output		= wan_iface_output;
	dev->if_start		= NULL;	/*wan_iface_start;*/
	dev->if_watchdog	= wan_iface_watchdog;
	dev->if_snd.ifq_len	= 0;
	WAN_IFQ_SET_MAXLEN(&dev->if_snd, 100);

	IFATTACH(dev);
	if (!is_netdev){
		DEBUG_EVENT("%s: Attaching network interface to SPPP device!\n",
					wan_netif_name(dev));
		WAN_SPPP_ATTACH(dev);
		BPF_ATTACH(dev, DLT_PPP);
	}else{
		BPF_ATTACH(dev, DLT_RAW);
	}

	return 0;
}

static void wan_iface_detach (netdevice_t* dev, int is_netdev)
{
	DEBUG_EVENT("%s: Unregister interface!\n", 
			wan_netif_name(dev));
	if (!is_netdev){
		WAN_SPPP_DETACH(dev);
	}
	if (ifunit(wan_netif_name(dev))){
		IFDETACH(dev);
	}
	wan_netif_set_priv(dev, NULL);
}


static int
wan_iface_output(netdevice_t *dev, netskb_t *skb, struct sockaddr *dst, struct rtentry *rt0)
{
	wanpipe_common_t	*common;
	wan_spinlock_t		s;
	int 			kernel_entry = 0, err = 0;

	common = wan_netif_priv(dev);
	WAN_ASSERT(common == NULL);
	if (common->iface.output == NULL){
		DEBUG_EVENT("%s: Protocol output function is undefined!\n",
				wan_netif_name(dev));
		IF_DROP(&dev->if_snd);
		wan_skb_free(skb);
		return -EINVAL;
	}

	if (!wan_skb_mark(skb)){
		kernel_entry = 1;
	}
#if 0
	if (wan_skb_check(skb)){
#endif
		if (kernel_entry && wan_skb2buffer((void**)&skb)){
			DEBUG_EVENT("%s: Failed to correct mbuf list!\n",
				wan_netif_name(dev));
			IF_DROP(&dev->if_snd);
			wan_skb_free(skb);
			return ENOBUFS;
		}
#if 0
	}
#endif

	DEBUG_TEST("%s: Sending new packet %d bytes...\n",
			wan_netif_name(dev), wan_skb_len(skb));
	wan_bpf_report(dev, skb, 1);
	wan_spin_lock(&s);
	err = common->iface.output(dev, skb, dst, rt0);
	wan_spin_unlock(&s);
	if (err == 1 && kernel_entry){
		IF_DROP(&dev->if_snd);
		wan_skb_free(skb);
		err = ENETDOWN;
	}
	return err;
}


static int wan_iface_ioctl(netdevice_t* dev, u_long cmd, caddr_t data)
{
	wanpipe_common_t*	common;
	struct if_settings	ifsettings;
	int			err = 0, s;

#if defined(__NetBSD__) && (__NetBSD_Version__ >= 106000200)
	s = splvm();
#else
	s = splimp();
#endif
	common = wan_netif_priv(dev);
	switch(cmd){
	case SIOCSIFADDR:
		if (!common->is_netdev){
			err = 1;
		}
		break;

	case SIOCSIFFLAGS:
	    	/*
     		** If the interface is marked up - enable communications. 
	     	** If down - disable communications.  IFF_UP is taken 
		** care of before entering this function.
	     	*/
    		if ((dev->if_flags & IFF_UP) == 0){ 
			/* bring it down */
			DEBUG_EVENT("%s: Bringing interface down.\n", 
					wan_netif_name(dev));
			if (common->iface.close){
				common->iface.close((netdevice_t*)dev);
			}
			dev->if_flags &= ~IFF_RUNNING;
		}else{ /* bring it up */ 
			DEBUG_EVENT("%s: Bringing interface up.\n", 
					wan_netif_name(dev));
			if (common->iface.open){
				common->iface.open((netdevice_t*)dev);
			}
			dev->if_flags |= IFF_RUNNING;
		}
		break;

	case SIOC_WANPIPE_DEVICE:
		err = WAN_COPY_FROM_USER(&ifsettings, 
					((struct ifreq*)data)->ifr_data,
					 sizeof(struct if_settings));
		if (err){
			DEBUG_EVENT("%s: Failed to copy from user space!\n",
						wan_netif_name(dev));
			goto wan_iface_ioctl_done;
		}
		switch(ifsettings.type){
		case IF_GET_PROTO:
			ifsettings.type = common->protocol;
			err = WAN_COPY_TO_USER(((struct ifreq*)data)->ifr_data,
						&ifsettings,
						sizeof(struct if_settings));
			if (err){
				DEBUG_EVENT("%s: Failed to copy to user space!\n",
							wan_netif_name(dev));
			}
			break;

		case IF_PROTO_CISCO:
		case IF_PROTO_PPP:
			err = wan_iface_set_proto(dev, (struct ifreq*)data);
			break;	
		
		default:
			if (common->iface.ioctl){
				/* Argument seqeunce is change for Linux order */
				err = common->iface.ioctl(dev, (struct ifreq*)data, cmd);
			}
			break;
			
		}
		break;

	default:
		if (common->iface.ioctl){
			/* Argument seqeunce is change for Linux order */
			err = common->iface.ioctl(dev, (struct ifreq*)data, cmd);
		}
		break;
	}

	if (err && !common->is_netdev){
		err = WAN_SPPP_IOCTL(dev, cmd, data); 
	}
wan_iface_ioctl_done:
	splx(s);
	return err;
}

#if defined(__FreeBSD__)
static void wan_iface_init(void* dev)
{
	return;
}
#endif

static void wan_iface_watchdog(netdevice_t* dev)
{
	return;
}

static int wan_iface_input(netdevice_t* dev, netskb_t* skb)
{
#if defined(NETGRAPH)
	sdla_t			*card;
#endif
	wanpipe_common_t	*common;

	if ((common = wan_netif_priv(dev)) == NULL){
		return -EINVAL;
	}

	skb->m_pkthdr.rcvif = dev;
	wan_bpf_report(dev, skb, 1);
#if defined(NETGRAPH)
	if ((card = common->card) == NULL){
		return -EINVAL;
	}
	if (card->hook){
		int error;
		NG_SEND_DATA_ONLY(error, 
				  card->hook, 
				  skb);
		card->inbytes += wan_skb_len(skb);
	}
#else
	/* If we are sending it up, remove our M_FLAG */
	wan_skb_clear_mark(skb);

	if (common->is_netdev){
# if defined(__FreeBSD__) && (__FreeBSD_version >= 501000)
		netisr_queue(
			(skb->m_flags & M_PROTO1)?NETISR_IPX:NETISR_IP, skb); 
# else
		struct ifqueue	*intrq = NULL;
		int		intr_type = 0;
		wan_spinlock_t s;
	
		wan_spin_lock(&s);
		if (skb->m_flags & M_PROTO1){
#ifdef IPX
			intrq = &ipxintrq; 
			intr_type = NETISR_IPX;
#else
			DEBUG_EVENT("%s: IPX protocol is notenabled!!\n",
						wan_netif_name(dev));
			return -EINVAL;
#endif
		}else{
			intrq = &ipintrq; 
			intr_type = NETISR_IP;
		}
		schednetisr(intr_type);
		if (IF_QFULL(intrq)){
			IF_DROP(intrq);
#if 0
			/* Do not free if we are returning non zero value */
			wan_skb_free(skb);	/*m_freem(ipkt);*/
#endif
			wan_spin_unlock(&s);
			return -EINVAL;
		}

		/* ok enqueue the packet */
		IF_ENQUEUE(intrq, (netskb_t*)skb);
		wan_spin_unlock(&s);
# endif
	}else{
		WAN_SPPP_INPUT(dev, skb);
	}
#endif
	return 0;
}

static int wan_iface_set_proto(netdevice_t* dev, struct ifreq* ifr)
{
	wanpipe_common_t*	common;
	struct if_settings*	ifsettings;
	int			err = 0;

	if ((common = wan_netif_priv(dev)) == NULL){
		DEBUG_EVENT("%s: Private structure is null!\n",
				wan_netif_name(dev));
		return -EINVAL;
	}
	if (common->is_netdev){
		DEBUG_EVENT("%s: Invalid ioctl call!\n",
				wan_netif_name(dev));
		return -EINVAL;
	}
	ifsettings = (struct if_settings*)ifr->ifr_data;
	switch (ifsettings->type) {
	case IF_PROTO_CISCO:
		common->protocol = WANCONFIG_CHDLC;
#if defined(__FreeBSD__)
		dev->if_flags |= IFF_LINK2;
#elif defined(__OpenBSD__)
		((struct sppp *)dev)->pp_flags |= PP_CISCO;
#endif
		((struct sppp *)dev)->pp_flags |= PP_KEEPALIVE;
		break;
	case IF_PROTO_PPP:
		common->protocol = WANCONFIG_PPP;
#if defined(__FreeBSD__)
		dev->if_flags &= ~IFF_LINK2;
#elif defined(__OpenBSD__)
		((struct sppp *)dev)->pp_flags &= ~PP_CISCO;
#endif
		((struct sppp *)dev)->pp_flags &= ~PP_KEEPALIVE;
		break;
	}
	err = WAN_SPPP_IOCTL(dev, SIOCSIFFLAGS, ifr);
	return err;  
}

/************************************ END **********************************/
